//
// Synchronous FIFO 
// Output port is half the width of the input port
//
//
// Freeware September 2015, Fen Logic Ltd.
// This code is free and is delivered 'as is'.
// It comes without any warranty, to the extent permitted by applicable law.
// There are no restrictions to any use or re-use of this code
// in any form or shape. It would be nice if you keep my company name
// in the source or modified source code but even that is not
// required as I can't check it anyway.
// But the code comes with no warranties or guarantees whatsoever.
//
// Read and Write are ignored when empty/full
// The fifo level is registered
// The full, empty flags can be registered or not
//
//
// Known issues:
//    artithmetic is not size compliant
//    Storage contents is not cleared
//    relying on synthesis tool to remove unused logic
//    If Depth is set to a power of two the read and write pointers
//       are one bit too big. Assuming synthesis will remove that.
//


module sync_fifo_hd
#(parameter WIDTH    =  8, // Unit width in bits
            DEPTH    = 48, // Output units depth (must be even)
            LEVLBITS =  6, // ciel( log2( Units DEPTH )) (+1)
            REGFLAGS =  1  // Full, empty are registered
)
(
   input                   clk,
   input                   reset_n,
   input                   enable, // clock gating
   input                   clear,

   input                   write,
   input     [WIDTH*2-1:0] wdata, // MS is entry[1], LS is entry[0]
   input                   read,
   output      [WIDTH-1:0] rdata,


   output reg                empty,
   output reg                full,
   output reg [LEVLBITS-1:0] level // Read data level 
                                   // For the write data level use level>>1
                                   
);

// Check LEVLBITS against DEPTH and check DEPTH is even
localparam AW = DEPTH >= (1<<LEVLBITS) ? 0 : // 0 means error
                DEPTH  &  1            ? 0 : // 0 means error
                LEVLBITS;

localparam [AW-1:0] ZERO = {AW{1'b0}}; // should give error on AW=0


reg [WIDTH*2-1:0] memory [0:DEPTH/2-1];
reg [     AW-1:0] rd_pntr; // Full readpointer
reg [     AW-2:0] wt_pntr; // One bit less write pointer
reg               full_r,empty_r;

wire l_read,l_write;

   // prevent disastrous reads & writes
   // by refusing read if empty or write if full
   assign l_read  = read  & !empty;
   assign l_write = write & !full;


   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         rd_pntr <= ZERO;
         wt_pntr <= ZERO;
         level   <= ZERO;
         full_r  <= 1'b0;
         empty_r <= 1'b1;
      end
      else if (enable)
      begin
         if (clear)
         begin
           rd_pntr <= ZERO;
           wt_pntr <= ZERO;
           level   <= ZERO;
           full_r  <= 1'b0;
           empty_r <= 1'b1;
         end
         else
         begin
            case ({l_read,l_write})
            2'b01 : // Write, no read
              begin
                 if (wt_pntr==DEPTH/2-1)
                    wt_pntr <= ZERO;
                 else
                    wt_pntr <= wt_pntr + 1;
                 level   <= level + 2;
                 full_r  <= (level>=DEPTH-3);
                 empty_r <= 1'b0;
              end

            2'b10 : // Read, no write
               begin
                  if (rd_pntr==DEPTH-1)
                     rd_pntr <= ZERO;
                  else
                     rd_pntr <= rd_pntr + 1;
                  level   <= level - 1;
                  full_r  <= (level>=DEPTH);
                  empty_r <= (level==1);
               end

            2'b11 : // Read & Write
               begin
                  // actual two writes and one read
                  if (wt_pntr==DEPTH/2-1)
                     wt_pntr <= ZERO;
                  else
                     wt_pntr <= wt_pntr + 1;

                  if (rd_pntr==DEPTH-1)
                     rd_pntr <= ZERO;
                  else
                     rd_pntr <= rd_pntr + 1;

                  level   <= level + 1;
                  full_r  <= (level>=DEPTH-2);
                  empty_r <= 1'b0;
               end
            endcase
        end

`ifdef ASSERT_ON
         // Catch fatal operational errors
         if (write && !l_write)
         begin
            $display("%m,@%0t: write to full FIFO",$time);
            #1 $stop;
         end

         if (read && !l_read)
         begin
            $display("%m,@%0t: read from empty FIFO",$time);
            #1 $stop;
         end
`endif

      end
   end


wire [WIDTH*2-1:0] mrdata;
wire [     AW-2:0] read_adrs = rd_pntr>>1;

   // Write to fifo
   always @(posedge clk)
   begin
      if (enable & l_write)
         memory[wt_pntr] <= wdata;
   end

   // Read from fifo.
   // Use MS address bits to read entry
   // LS address bit selects upper/lower
   assign mrdata = memory[ read_adrs ];
   assign rdata = rd_pntr[0] ? mrdata[WIDTH*2-1 : WIDTH] : mrdata[WIDTH-1 : 0];
   

   // empty and full flags depend on parameter
   // registered or not
   always @( * )
   begin
      if (REGFLAGS==0)
      begin
         full  = (level>=DEPTH-1);
         empty = (level==0);
      end
      else
      begin
        full  = full_r;
        empty = empty_r;
      end
   end


endmodule